/*
 * Library: dxflib.r
 * Author: Vasilis.Vlachoudis@cern.ch 2005
 */

/* --- DXFPrimitives --- */
DXFPrimitives: procedure expose prim. verbose vecXo vecYo vecZo
	parse arg fout

	/* --- global variables --- */
	spherefaces = 16
	cylfaces    = 24
	planesize   = 10.0
	infcylsize  = 10.0
	PI          = 3.141592654

	primitives = "RPP BOX SPH RCC REC TRC ELL WED ARB XYP XZP YZP PLA XCC YCC ZCC XEC YEC ZEC"
	colors     = "  5  10  15  20  25  30  35  40  45  50  55  60  65  70  75  80  85  90  95"

	/* --- 1st pass, scan geometry and write the TABLE section --- */
	call DXFWrite fout,0,"SECTION"
	call DXFWrite fout,2,"TABLES"
	call DXFTableLTYPE fout

	call DXFWrite fout,0,"TABLE"
	call DXFWrite fout,2,"LAYER"
	call DXFWrite fout,70,prim.0

	do i=1 to prim.0
		call DXFWrite fout,0,"LAYER"
		call DXFWrite fout,2,DXFObjName(prim.i.@prim,prim.i,i)
		call DXFWrite fout,70,0
		/*call DXFWrite fout,62,DXFObjColor(prim.i.@prim)*/
		call DXFWrite fout,62,i
		call DXFWrite fout,6,"CONTINUOUS"
	end
	call DXFWrite fout,0,"ENDTAB"
	call DXFTableSTYLE fout
	call DXFTableUCS fout
	call DXFWrite fout,0,"ENDSEC"

	call DXFWrite fout,0,"SECTION"
	call DXFWrite fout,2,"ENTITIES"

	do n=1 to prim.0
		name      = prim.n
		primitive = prim.n.@prim
		id        = DXFObjName(primitive,name,n)
		color     = DXFObjColor(primitive)
		color     = n

		if verbose then say n id primitive name
		select
			when primitive="RPP" then call DXFRectangular
			when primitive="BOX" then call DXFBox
			when primitive="SPH" then call DXFSphere
			when primitive="RCC" then call DXFCylinder
			when primitive="REC" then call DXFEllipticalCylinder
			when primitive="TRC" then call DXFTruncatedCone
			/***
			when primitive="ELL" then call DXFEllipsoid
			when primitive="WED" then call DXFAngleWedge
			when primitive="ARB" then call DXFPolyhedron
			when primitive="ARB" then call DXFPolyhedron
			***/
			when primitive="XYP" then call DXFPlane 'Z'
			when primitive="XZP" then call DXFPlane 'Y'
			when primitive="YZP" then call DXFPlane 'X'
			when primitive="PLA" then call DXFPlane 'A'
			when primitive="XCC" then call DXFInfCylinder 'X'
			when primitive="YCC" then call DXFInfCylinder 'Y'
			when primitive="ZCC" then call DXFInfCylinder 'Z'
			when primitive="XEC" then call DXFInfEllipticalCylinder 'X'
			when primitive="YEC" then call DXFInfEllipticalCylinder 'Y'
			when primitive="ZEC" then call DXFInfEllipticalCylinder 'Z'
			otherwise
				say "ERROR: unknown body type" primitive
				leave
		end
	end

	call DXFWrite fout,0,"ENDSEC"
	call DXFWrite fout,0,"EOF"
return

/* ------------- DXFObjName ------------ */
DXFObjName: procedure
	parse arg prim,name,id
	if datatype(name,"N") then
		return right(name,4,0)"_"prim
	else
		return name

/* ------------- DXFObjColor ------------ */
DXFObjColor: procedure expose primitives colors
	parse arg prim "." .
	p = wordpos(prim,primitives)
	if p=0 then return 1
return word(colors,p)

/* ---------------- DXFWriteHeader ------------- */
DXFWriteHeader: procedure
	parse arg f
	call DXFWrite f,0,"SECTION"
	call DXFWrite f,2,"HEADER"

	call DXFWrite f,9,"$ACADVER"
	call DXFWrite f,1,"AC1008"

	call DXFWrite f,9,"$UCSORG"
	call DXFWriteVec f,0.0 0.0 0.0

	call DXFWrite f,9,"$UCSXDIR"
	call DXFWriteVec f,1.0 0.0 0.0

	call DXFWrite f,9,"$TILEMODE"
	call DXFWrite f,70,1

	call DXFWrite f,9,"$UCSYDIR"
	call DXFWriteVec f,0.0 1.0 0.0

	call DXFWrite f,0,"ENDSEC"
return

/* ---------------- DXFTableLTYPE ------------- */
DXFTableLTYPE: procedure
	parse arg f
	call DXFWrite f,0,"TABLE"
	call DXFWrite f,2,"LTYPE"
	call DXFWrite f,70,1
	call DXFWrite f,0,"LTYPE"
	call DXFWrite f,2,"CONTINUOUS"
	call DXFWrite f,70,64
	call DXFWrite f,3,"Solid line"
	call DXFWrite f,72,65
	call DXFWrite f,73,0
	call DXFWrite f,40,0.0
	call DXFWrite f,0,"ENDTAB"
return

/* ---------------- DXFTableSTYLE ------------- */
DXFTableSTYLE: procedure
	parse arg f
	call DXFWrite f,0,"TABLE"
	call DXFWrite f,2,"STYLE"
	call DXFWrite f,70,1
	call DXFWrite f,0,"STYLE"
	call DXFWrite f,2,"STANDARD"
	call DXFWrite f,70,0
	call DXFWrite f,40,0.0
	call DXFWrite f,41,1.0
	call DXFWrite f,50,0.0
	call DXFWrite f,71,0
	call DXFWrite f,42,0.2
	call DXFWrite f,3,"txt"
	call DXFWrite f,4,""
	call DXFWrite f,0,"ENDTAB"
return

/* ---------------- DXFTableUCS ------------- */
DXFTableUCS: procedure
	parse arg f
	call DXFWrite f,0,"TABLE"
	call DXFWrite f,2,"UCS"
	call DXFWrite f,70,0
	call DXFWrite f,0,"ENDTAB"
return

/* -------- DXFPolyBegin --------- */
DXFPolyBegin: procedure
	parse arg f,name,color,vertices,faces
	call DXFWrite f,0,"POLYLINE"
	call DXFWrite f,8,name
	call DXFWrite f,66,1
	call DXFWrite f,70,64
	call DXFWrite f,71,vertices
	call DXFWrite f,72,faces
	call DXFWrite f,62,color
return

/* -------- DXFPolyEnd --------- */
DXFPolyEnd: procedure
	parse arg f,name
	call DXFWrite f,0,"SEQEND"
	call DXFWrite f,8,name
return

/* -------- DXFVertex --------- */
DXFVertex: procedure
	parse arg f,name,vec
	call DXFWrite f,0,"VERTEX"
	call DXFWrite f,8,name
	call DXFWriteVec f,vec
	call DXFWrite f,70,192
return

/* -------- DXFFace -------- */
DXFFace: procedure
	parse arg f,name,a,b,c,d
	call DXFWrite f,0,"VERTEX"
	call DXFWrite f,8,name
	call DXFWriteVec f,0 0 0
	call DXFWrite f,70,128
	if a^="" then call DXFWrite f,71,a
	if b^="" then call DXFWrite f,72,b
	if c^="" then call DXFWrite f,73,c
	if d^="" then call DXFWrite f,74,d
return

/* -------- DXFWriteVec ----------- */
DXFWriteVec: procedure
	parse arg f,x y z
	call DXFWrite f,10,x
	call DXFWrite f,20,y
	call DXFWrite f,30,z
return

/* -------- DXFWrite ------------ */
DXFWrite: procedure
	parse arg f,group,value
	call lineout f,group
	call lineout f,value
return

/* --- DXFRectangular --- */
DXFRectangular:
	call DXFPolyBegin fout,id,color,8,12

	call DXFVertex fout,id,VecMake(prim.n.@xmin, prim.n.@ymin, prim.n.@zmax)
	call DXFVertex fout,id,VecMake(prim.n.@xmax, prim.n.@ymin, prim.n.@zmax)
	call DXFVertex fout,id,VecMake(prim.n.@xmax, prim.n.@ymax, prim.n.@zmax)
	call DXFVertex fout,id,VecMake(prim.n.@xmin, prim.n.@ymax, prim.n.@zmax)

	call DXFVertex fout,id,VecMake(prim.n.@xmin, prim.n.@ymin, prim.n.@zmin)
	call DXFVertex fout,id,VecMake(prim.n.@xmax, prim.n.@ymin, prim.n.@zmin)
	call DXFVertex fout,id,VecMake(prim.n.@xmax, prim.n.@ymax, prim.n.@zmin)
	call DXFVertex fout,id,VecMake(prim.n.@xmin, prim.n.@ymax, prim.n.@zmin)

	call DXFFace fout,id, 1, 2,-3
	call DXFFace fout,id,-1, 3, 4
	call DXFFace fout,id, 1, 5,-6
	call DXFFace fout,id,-1, 6, 2
	call DXFFace fout,id, 2, 6,-7
	call DXFFace fout,id,-2, 7, 3
	call DXFFace fout,id, 3, 7,-8
	call DXFFace fout,id,-3, 8, 4
	call DXFFace fout,id, 4, 8,-5
	call DXFFace fout,id,-4, 5, 1
	call DXFFace fout,id, 5, 8,-7
	call DXFFace fout,id,-5, 7, 6

	call DXFPolyEnd fout,id
return

/* ---------- DXFBox -------- */
DXFBox:
	r = prim.n.@P
	a = prim.n.@X
	b = prim.n.@Y
	c = prim.n.@Z

	call DXFPolyBegin fout,id,color,8,12

	call DXFVertex fout,id,VecAdd(r,c)
	call DXFVertex fout,id,VecAdd(VecAdd(r,a),c)
	call DXFVertex fout,id,VecAdd(VecAdd(VecAdd(r,a),b),c)
	call DXFVertex fout,id,VecAdd(VecAdd(r,b),c)

	call DXFVertex fout,id,r
	call DXFVertex fout,id,VecAdd(r,a)
	call DXFVertex fout,id,VecAdd(VecAdd(r,a),b)
	call DXFVertex fout,id,VecAdd(r,b)

	call DXFFace fout,id, 1, 2,-3
	call DXFFace fout,id,-1, 3, 4
	call DXFFace fout,id, 1, 5,-6
	call DXFFace fout,id,-1, 6, 2
	call DXFFace fout,id, 2, 6,-7
	call DXFFace fout,id,-2, 7, 3
	call DXFFace fout,id, 3, 7,-8
	call DXFFace fout,id,-3, 8, 4
	call DXFFace fout,id, 4, 8,-5
	call DXFFace fout,id,-4, 5, 1
	call DXFFace fout,id, 5, 8,-7
	call DXFFace fout,id,-5, 7, 6

	call DXFPolyEnd fout,id
return

/* --- DXFSphere --- */
DXFSphere:
	center = prim.n.@P
	radius = prim.n.@R

	isim = spherefaces % 2
	if isim//2 =0 then isim = isim - 1

	vertex=1

	call DXFPolyBegin fout,id,color, isim*spherefaces+2, 2*spherefaces*isim

	stepFi = 2*PI / spherefaces
	step0  = PI / (isim+1)

	angle0=step0
	do j=1 to isim
		angleFi = 0
		do i=1 to sphereFaces
			angleFi = angleFi + stepFi
			v = VecMake(	radius * cos(angleFi) * sin(angle0),,
					radius * sin(angleFi) * sin(angle0),,
					radius * cos(angle0) )
			call DXFVertex fout,id,VecAdd(v,center)
			vertex = vertex + 1
		end
		angle0 = angle0 + step0
	end

	downvertex = vertex
	upvertex = downvertex+1

	call DXFVertex fout,id,VecAdd(center,VecMake(0,0,-radius))
	call DXFVertex fout,id,VecAdd(center,VecMake(0,0, radius))

	js = 0
	do j=0 to isim-2
		do i=0 to spherefaces-1
			a = i+js+1
			b = (i+1)//spherefaces+spherefaces+js+1
			c = (i+1)//spherefaces+js+1
			call DXFFace fout,id, -a,b,c

			b = i+spherefaces+js+1
			c = (i+1)//spherefaces+spherefaces+js+1
			call DXFFace fout,id, -a,-b,-c
		end
		js = js + spherefaces
	end

	do i=0 to spherefaces-1
		call DXFFace fout,id, ((i+1)//spherefaces+1), -upvertex, -(i+1)
	end
	js = (isim-1) * spherefaces
	do i=0 to spherefaces-1
		call DXFFace fout,id, (i+js+1), -downvertex, ((i+1)//spherefaces+js+1)
	end
	call DXFPolyEnd fout,id
return

/* --------- Elliptical Cylinder --------- */
DXFEllCyl:
	parse arg id, r, x, y, z, rx, ry

	call DXFPolyBegin fout,id,color, 2*cylfaces+2, 4*cylfaces

	stepangle = 2*PI/cylfaces

	/* first make the top part */
	angle = 0
	do i=1 to cylfaces
		v = VecComb(rx*cos(angle), x, ry*sin(angle), y)
		call DXFVertex fout,id,VecAdd(VecAdd(v,z),r)
		angle = angle + stepangle
	end

	/* and then the base */
	angle = 0
	do i=1 to cylfaces
		v = VecComb(rx*cos(angle), x, ry*sin(angle), y)
		call DXFVertex fout,id,VecAdd(v,r)
		angle = angle + stepangle
	end

	call DXFVertex fout,id,VecAdd(r,z)
	call DXFVertex fout,id,r

	do i=1 to cylfaces
		a = i
		b = i+cylfaces
		c = i//cylfaces+1
		call DXFFace fout,id, a,-b,-c

		a = c
		c = i//cylfaces+cylfaces+1
		call DXFFace fout,id, -a,-b,-c
	end

	b = 2*cylfaces+1
	do i=1 to cylfaces
		a = i
		c = i//cylfaces+1
		call DXFFace fout,id, -a,c,-b
	end

	b = 2*cylfaces+2
	do i=1 to cylfaces
		a = i+cylfaces
		c = i//cylfaces+cylfaces+1
		call DXFFace fout,id, -a,-b,c
	end

	call DXFPolyEnd fout,id
return

/* --- DXFCylinder --- */
DXFCylinder:
	r = prim.n.@P
	z = prim.n.@Z
	radius = prim.n.@R

	/* find two axis perpendicular to height */
	/* Let's try Z-axis */
	v = VecNorm(z)
	if abs(VecDot(v,VecZo)) > 0.9
		then x = VecXo	/* use another axis */
		else x = VecZo

	/* find y from the cross product */
	y = VecCross(z,x)
	/* find the correct x with the cross product */
	x = VecCross(y,z)

	/* normalise them */
	x = VecNorm(x)
	y = VecNorm(y)

	call DXFEllCyl id,r,x,y,z,radius,radius
return

/* --- DXFInfCylinder --- */
DXFInfCylinder:
	parse arg axis

	R      = VecAddS(-infcylsize/2,prim.n.@N,prim.n.@P)
	Z      = VecS(infcylsize,prim.n.@N)
	radius = prim.n.@R

	/* find two axis perpendicular to height */
	/* Let's try Z-axis */
	v = VecNorm(z)
	if abs(VecDot(v,VecZo)) > 0.9
		then x = VecXo	/* use another axis */
		else x = VecZo

	/* find y from the cross product */
	y = VecCross(z,x)
	/* find the correct x with the cross product */
	x = VecCross(y,z)

	/* normalise them */
	x = VecNorm(x)
	y = VecNorm(y)

	call DXFEllCyl id,r,x,y,z,radius,radius
	call DXFPolyEnd fout,id
return

/* --- DXFInfEllipticalCylinder --- */
DXFInfEllipticalCylinder:
	parse arg axis

	R      = VecAddS(-infcylsize/2,prim.n.@N,prim.n.@P)
	Z      = VecS(infcylsize,prim.n.@N)
	select
		when axis=='X' then parse var prim.n.@R . r1 r2
		when axis=='Y' then parse var prim.n.@R r2 . r1
		when axis=='Z' then parse var prim.n.@R r1 r2 .
	end
	X      = VecNorm(prim.n.@X)
	Y      = VecNorm(prim.n.@Y)
	call DXFEllCyl id,r,x,y,z,r1,r2
return

/* --------- DXFEllipticalCylinder --------- */
DXFEllipticalCylinder:
	r = prim.n.@P
	z = prim.n.@Z
	x = prim.n.@X
	y = prim.n.@Y

	rx = VecLen(x)
	ry = VecLen(y)

	/* normalise them */
	x = VecNorm(x)
	y = VecNorm(y)

	call DXFEllCyl id,r,x,y,z,rx,ry
return

/* --- DXFTruncatedCone --- */
DXFTruncatedCone:
	r = prim.n.@P
	z = prim.n.@Z
	baseRadius = prim.n.@bR
	apexRadius = prim.n.@aR

	/* find two axis perpendicular to height */
	/* Let's try Z-axis */
	v = VecNorm(z)
	if abs(VecDot(v,VecZo)) > 0.9
		then x = VecXo	/* use another axis */
		else x = VecZo

	/* find y from the cross product */
	y = VecCross(z,x)
	/* find the correct x with the cross product */
	x = VecCross(y,z)

	/* normalise them */
	x = VecNorm(x)
	y = VecNorm(y)

	call DXFPolyBegin fout,id,color, 2*cylfaces+2, 4*cylfaces

	stepangle = 2*PI/cylfaces

	/* first make the top part */
	angle = 0
	do i=1 to cylfaces
		v = VecComb(apexRadius*cos(angle), x, apexRadius*sin(angle), y)
		call DXFVertex fout,id,VecAdd(VecAdd(v,z),r)
		angle = angle + stepangle
	end

	/* and then the base */
	angle = 0
	do i=1 to cylfaces
		v = VecComb(baseRadius*cos(angle), x, baseRadius*sin(angle), y)
		call DXFVertex fout,id,VecAdd(v,r)
		angle = angle + stepangle
	end

	call DXFVertex fout,id,VecAdd(r,z)
	call DXFVertex fout,id,r

	do i=1 to cylfaces
		a = i
		b = i+cylfaces
		c = i//cylfaces+1
		call DXFFace fout,id, a,-b,-c

		a = c
		c = i//cylfaces+cylfaces+1
		call DXFFace fout,id, -a,-b,-c
	end

	b = 2*cylfaces+1
	do i=1 to cylfaces
		a = i
		c = i//cylfaces+1
		call DXFFace fout,id, -a,-b,c
	end

	b = 2*cylfaces+2
	do i=1 to cylfaces
		a = i+cylfaces
		c = i//cylfaces+cylfaces+1
		call DXFFace fout,id, -a,-b,c
	end

	call DXFPolyEnd fout,id
return

/* --- DXFPlane --- */
DXFPlane:
	z = prim.n.@N
	r = prim.n.@P

	/* find two axis perpendicular to normal z */
	/* Let's try Z-axis */
	if abs(VecDot(z,VecZo)) > 0.9
		then x = VecXo	/* use another axis */
		else x = VecZo

	/* find y from the cross product */
	y = VecCross(z,x)
	/* find the correct x with the cross product */
	x = VecCross(y,z)

	/* normalise them */
	x = VecNorm(x)
	y = VecNorm(y)

	call DXFPolyBegin fout,id,color, 4, 4

	call DXFVertex fout,id,VecAdd(r,VecComb( planesize,x, planesize,y))
	call DXFVertex fout,id,VecAdd(r,VecComb(-planesize,x, planesize,y))
	call DXFVertex fout,id,VecAdd(r,VecComb(-planesize,x,-planesize,y))
	call DXFVertex fout,id,VecAdd(r,VecComb( planesize,x,-planesize,y))

	call DXFFace fout,id, 1, 2, 3
	call DXFFace fout,id, 1, 3, 4
	call DXFFace fout,id, 1, 4, 2
	call DXFFace fout,id, 4, 3, 2

	call DXFPolyEnd fout,id
return
