#!/bin/env rexx

trace o
call import "files.r"
call import "pslib.r"
call import "veclib.r"
call import "flukalib.r"

signal on halt
signal on syntax
signal on error

numeric digits 10

parse arg inputfile flukaexe .
if inputfile = "" then do
	say "syntax: flukaplot.r <input-file> [<flukaexe>]"
	exit
end
if flukaexe<>"" & ^exist(flukaexe) then do
	say "+++ Fluka executable '"flukaexe"' not found."
	exit 1
end

tmpfile = "flplt"date("s")"."time("s")
fin = open(inputfile,"r")
if fin<0 then do
	say "ERROR: Opening file" inputfile
	exit 1
end

fout = open(tmpfile".inp","w")
flukatitle = ""
do forever
	line = read(fin)
	if eof(fin) then leave
	card = strip(left(line,10))
	if card="TITLE" then do
		call lineout fout,line
		line = read(fin)
	end
	if card="START" | card="STOP" then line = "*"line
	else
	if card="PLOTGEOM" then do
		line = "*"line
		do 4
			call read fin
		end
		iterate
	end
	call lineout fout,line
end
call close fin
call close fout

/* Create a pipe to ghostscript */
/*fps = gsinit(tmpfile".pipe","x11","-g1011x781")*/
/*fps = gsinit(tmpfile".pipe","x11","-g780x650 -r72x72")*/
fps = gsinit(tmpfile".pipe","x11","-g792x612 -r72x72")

FLUKA_PATH="${FLUPRO}/flutil"

if flukaexe<>"" then
	rfluka = FLUKA_PATH"/rfluka -M1 -e" flukaexe
else
	rfluka = FLUKA_PATH"/rfluka -M1"

/* */
type	   = "M"
xx	   = 0
yy	   = 0
zz	   = 0
extX	   = 50
extY	   = 50
axisX	   = "1.0 0.0 0.0"
axisY	   = "0.0 1.0 0.0"
deffile    = "flukaplot.ps"
arrowscale = 1.0

datafile   = ""
binno	   = 1
norm	   = ""
projaxis   = "Z"
xmin	   = ""
ymin	   = ""
zmin	   = ""
xmax	   = ""
ymax	   = ""
zmax	   = ""
plotmin    = ""
plotmax    = ""
divpd	   = 3

first	   = 1

updateplot = 0
movefile = ""
do forever
	showPlotgeom = 0
	say "plot>"
	parse pull line
	line = translate(line,,",")	/* Get rid of commas */
	do while line<>""
		parse var line cmd line
		upper cmd
		select
			when cmd="?" | abbrev("HELP",cmd) then do
				call Help
				iterate
			end
			when cmd="PX" then do
				parse var line xx line
				xx = toNum(xx)
				axisX = "0.0 0.0 1.0"
				axisY = "0.0 1.0 0.0"
				updateplot = 1
			end
			when cmd="PY" then do
				parse var line yy line
				yy = toNum(yy)
				axisX = "0.0 0.0 1.0"
				axisY = "1.0 0.0 0.0"
				updateplot = 1
			end
			when cmd="PZ" then do
				parse var line zz line
				zz = toNum(zz)
				axisX = "1.0 0.0 0.0"
				axisY = "0.0 1.0 0.0"
				updateplot = 1
			end
			when abbrev("CENTER",cmd) then do
				parse var line tx ty line
				tx = toNum(tx)
				ty = toNum(ty)
				if datatype(tx,"N") & datatype(ty,"N") then do
					v = VecAddS(tx,axisX,VecAddS(ty,axisY,xx yy zz))
					parse var v xx yy zz
					updateplot = 1
				end; else
					say "+++ Error in CENTER command"
			end
			when abbrev("ORIGIN",cmd) then do
				parse var line xx yy zz line
				xx = toNum(xx)
				yy = toNum(yy)
				zz = toNum(zz)
				updateplot = 1
			end
			when abbrev("FILE",cmd,2) then do
				parse var line movefile line
			end
			when abbrev("FACTOR",cmd,2) then do
				parse var line factor line
				if datatype(factor,"N") then do
					extX = extX/factor
					extY = extY/factor
					updateplot = 1
				end; else
					say "+++ Invalid factor" factor
			end
			when abbrev("BASIS",cmd) then do
				parse var line axx axy axz ayx ayy ayz line
				if datatype(axx,"N") & datatype(axy,"N") &,
				   datatype(axz,"N") & datatype(ayx,"N") &,
				   datatype(ayy,"N") & datatype(ayz,"N") then do
					axisX = VecNorm(axx axy axz)
					axisY = VecNorm(ayx ayy ayz)
					updateplot = 1
				end; else
					say "+++ Error on BASIS definition"
			end
			when abbrev("EXTEND",cmd,2) then do
				parse var line extx exty line
				if ^datatype(exty,"N") then do
					line = exty line
					exty = extx
				end
				updateplot = 1
			end
			when abbrev("DATA",cmd,2) then do
				parse var line datafile line
				if ^exist(datafile) then say "+++ ERROR file" datafile " do not exist"
				updateplot = 1
			end
			when abbrev("BIN",cmd,2) then do
				parse var line binno line
				if ^datatype(binno,"N") then say "+++ ERROR in binning number"
				updateplot = 1
			end
			when abbrev("LIMITS",cmd) then do
				parse var line axis min max line
				axis=translate(axis)
				if pos(axis,"XYZ")=0 then say "+++ ERROR in limits"
				if ^datatype(min,"N") then min=""
				if ^datatype(max,"N") then max=""
				select
					when axis="X" then do
						xmin = min
						xmax = max
						updateplot = 1
					end
					when axis="Y" then do
						ymin = min
						ymax = max
						updateplot = 1
					end
					when axis="Z" then do
						zmin = min
						zmax = max
						updateplot = 1
					end
				end
			end
			when abbrev("NORM",cmd) then do
				parse var line norm line
				interpret "norm="norm
				if ^datatype(norm,"N") then
					say "+++ ERROR in normalization factor"
				updateplot = 1
			end
			when abbrev("DIV",cmd,2) then do
				parse var line divpd line
				if ^datatype(divpd,"N") then do
					line = divpd line
					divpd = 3
				end
				updateplot = 1
			end
			when abbrev("MIN",cmd,2) then do
				parse var line plotmin line
				if ^datatype(plotmin,"N") then do
					line = plotmin line
					plotmin = ""
				end
				updateplot = 1
			end
			when abbrev("MAX",cmd,2) then do
				parse var line plotmax line
				if ^datatype(plotmax,"N") then do
					line = plotmax line
					plotmax = ""
				end
				updateplot = 1
			end
			when abbrev("PLOTGEOM",cmd,2) then showPlotgeom = 1
			when abbrev("PROJECT",cmd,2) then do
				parse var line projaxis line
				projaxis=translate(projaxis)
				if pos(projaxis,"XYZ")=0 then do
					say "+++ ERROR in project"
					projaxis = "Z"
				end
				updateplot = 1
			end
			when abbrev("TYPE",cmd) then do
				parse var line t line
				t1 = translate(left(t,1))
				if pos(t1,"GRMLFDP")>0 then do
					type = t1
					if type="D" & translate(substr(t,2,1))="G" then type="DG"
					if type="G" & translate(substr(t,2,1))="D" then type="DG"
					if type="P" & translate(substr(t,2,1))="G" then type="PG"
					if type="G" & translate(substr(t,2,1))="P" then type="PG"
					parse var line s line
					s2 = toNum(s)
					if datatype(s2,"N") then
						arrowscale = s2
					else
						line = s line
				end; else
					line = t line
				updateplot = 1
			end
			when cmd="END" | cmd="EXIT" |,
				abbrev("QUIT",cmd) then signal halt
		end
	end
	if showPlotgeom then call PlotGeomCard
	if updateplot then do
		call plot
		if movefile<>"" then do
			say "Saving file to" movefile
			"cp -f" deffile movefile
			movefile = ""
		end
		updateplot = 0
	end
end

halt:
syntax:
error:
	/* cleanup */
	call gsclose
	/*
	call lineout fps,"quit"
	call flush fps
	call close fps
	"rm" pipefile
	*/
	"rm -f" tmpfile"* ran"tmpfile"* pawlevh.* doslev.dat plotgeom.his"
exit

toNum: procedure
	x = translate(arg(1))
	if datatype(x,"N") then return x
return translate(x,"E","D")

/* --- Help --- */
Help:
	say
	say center(" Current Settings ",50,"-")
	say "Plot type.........." type
	say "Origin............." xx yy zz
	say "Extents............" extX extY
	say "Axis-X............." axisX
	say "Axis-Y............." axisY
	say "Output file........" deffile
	say "Arrow size........." arrowscale
	say "Data file.........." datafile
	say "Binning no........." binno
	if datatype(norm,"N") then
		invnorm="("1.0/norm")"
	else
		invnorm = ""
	say "Normalization......" norm invnorm
	say "Projection Axis...." projaxis
	say "Data X limits......" xmin xmax
	say ".... Y limits......" ymin ymax
	say ".... Z limits......" zmin zmax
	say "Plot limits........" plotmin plotmax
	say "Div. per decade...." divpd

	say
	say center(" Commands ",50,"-")
	say "Type GRMLFD [scale]	Plotting type"
	say "	G=Geometry	R=Regions	M=Materials"
	say "	L=Lattices	F=Mag.Field"
	say "	D=Data		DG=D+Geometry	P=R-Z-Phi"
	say "PX|PY|PZ n		plot a projection on X, Y or Z plane at position n"
	say "Center dx dy		move origin by (dx,dy) position on the projection plane"
	say "Origin x y z		move origin to (x,y,z)"
	say "FActor f		change extends by factor f"
	say "FIle filename		copy the current plot to file"
	say "Basis Xa Ya		plotting axis cosines"
	say "EXtend ex [ey]		plotting extends"
	say
	say "DAta datafile		USRBIN data file"
	say "BIn n			USRBIN binning number"
	say "Limits X|Y|Z min max	USRBIN plotting limits per axis"
	say "Norm n			USRBIN normalization"
	say "DIV n			Plotting divisions"
	say "MIn n|MAx n		Minimum and/or maximum of plotting histogram"
	say "PRoject X|Y|Z		USRBIN data projection axis"
	say
	say "Quit|EXIT|END		Terminate program"
	say
return

/* --- Perform the plot --- */
Plot:
	geomStore = tmpfile"001_PLOTGEOM.STORE"
	existPG = exist(geomStore)
	if left(type,1)="D" | left(type,1)=="P" then do
		if datafile="" then do
			say "+++ No datafile specified"
			return
		end
		if right(type,1)<>"G" then existPG=0

		/* Draw the data */
		if existPG then
			queue geomStore
		else
			queue "..."	/* Geometry */
		if projaxis="Z" then do
			queue "y"		/* Swap Axis */
			queue "2"		/* Bin data */
		end; else
			queue "n"

		if existPG then do
			queue "n"	/* Rotate geometry */
			queue ""
			queue ""
			queue ""
		end

		queue datafile
		if datatype(norm,"N") then
			queue -norm
		else
			queue ""
		queue ""
		queue binno
		queue xmin xmax
		if projaxis="X" then queue "1"; else queue ""
		queue ymin ymax
		if projaxis="Y" then queue "1"; else queue ""
		queue zmin zmax
		if projaxis="Z" then queue "1"; else queue ""
	end; else do
		/* Draw the geometry */
		if type="R" | type="M" then
			drawRegs = "1.0"
		else
			drawRegs = "0.0"
		line.1=left("PLOTGEOM",20)||right(drawRegs,10)
		line.2="Title"
		lowP  = VecAddS(-extX,axisX,VecAddS(-extY,axisY,xx yy zz))
		highP = VecAddS( extX,axisX,VecAddS( extY,axisY,xx yy zz))

		line.3 = right(FlukaFormat(VecX(lowP),10),10) ||,
			 right(FlukaFormat(VecY(lowP),10),10) ||,
			 right(FlukaFormat(VecZ(lowP),10),10) ||,
			 right(FlukaFormat(VecX(highP),10),10) ||,
			 right(FlukaFormat(VecY(highP),10),10) ||,
			 right(FlukaFormat(VecZ(highP),10),10)
		line.4 = right(FlukaFormat(VecX(axisY),10),10) ||,
			 right(FlukaFormat(VecY(axisY),10),10) ||,
			 right(FlukaFormat(VecZ(axisY),10),10) ||,
			 right(FlukaFormat(VecX(axisX),10),10) ||,
			 right(FlukaFormat(VecY(axisX),10),10) ||,
			 right(FlukaFormat(VecZ(axisX),10),10)

		line.5 = copies(right("1.0",10),2)||copies(right("300.0",10),2)
		line.6 = "STOP"

		if ^first then do
			queue "$-5,$d"
			queue "w"
			queue "q"
			"stack> ed -Gs" tmpfile".inp 2>&1 1>/dev/null"
		end
		first = 0

		do i=1 to 6
			queue line.i
		end
		"stack> cat >>" tmpfile".inp"

		"rm -f" tmpfile"001.err 2>&1 1>/dev/null"
		rfluka tmpfile "2>&1 1>/dev/null"

		if exist(tmpfile"001.err") then do
			say "+++ Geometry errors found"
		end
		plotgeom = tmpfile"001_PLOTGEOM.STORE"
		if rc<>0 | ^exist(plotgeom) then do
			say "+++ Error running fluka"
			return
		end

		queue plotgeom
	end

	do 10; queue ""; end
	"stack>" FLUKA_PATH"/pawlevbin (stack"

	do queued()
		parse pull tag det name
		if tag=="NB" & det=binno then title=name
	end

	/* Prepare paw plot */
	if extY>extX then do
		scaleX = extY/extX
		scaleY = ""
	end; else do
		scaleX = ""
		scaleY = extX/extY
	end

	queue "0"
	queue "exe $FLUPRO/flutil/bpawlev"
	queue ""
	queue deffile
	queue ""
	if left(type,1)="D" then
		queue title
	else
		queue flukatitle
		/*
		queue "Pos="||FlukaFormat(xx,10)","||,
				FlukaFormat(yy,10)","||,
				FlukaFormat(zz,10),
			"Ext="extX extY
		*/
	queue scaleX
	queue scaleY
	queue type
	select
		when type="F" then queue arrowscale
		when left(type,1)="D" then do
			queue plotmin
			queue plotmax
			queue divpd
		end
	end
	do 10; queue ""; end
	"stack> paw 2>&1 1>/dev/null"

	call UpdatePlot
return

/* --- PlotGeomCard --- */
PlotGeomCard:
	/* Draw the geometry */
	if type="R" then
		drawRegs = "1.0"
	else
		drawRegs = "0.0"
	line.1=left("PLOTGEOM",20)||right(drawRegs,10)||right("-2000.0",20)
	line.2="Title"
	lowP  = VecAddS(-extX,axisX,VecAddS(-extY,axisY,xx yy zz))
	highP = VecAddS( extX,axisX,VecAddS( extY,axisY,xx yy zz))

	line.3 = right(FlukaFormat(VecX(lowP),10),10) ||,
		 right(FlukaFormat(VecY(lowP),10),10) ||,
		 right(FlukaFormat(VecZ(lowP),10),10) ||,
		 right(FlukaFormat(VecX(highP),10),10) ||,
		 right(FlukaFormat(VecY(highP),10),10) ||,
		 right(FlukaFormat(VecZ(highP),10),10)
	line.4 = right(FlukaFormat(VecX(axisY),10),10) ||,
		 right(FlukaFormat(VecY(axisY),10),10) ||,
		 right(FlukaFormat(VecZ(axisY),10),10) ||,
		 right(FlukaFormat(VecX(axisX),10),10) ||,
		 right(FlukaFormat(VecY(axisX),10),10) ||,
		 right(FlukaFormat(VecZ(axisX),10),10)

	line.5 = copies(right("1.0",10),2)||copies(right("300.0",10),2)
	line.6 = "STOP"

	if showPlotgeom then do
		do i=1 to 5
			say "*" center(" PLOTGEOM ",70,"-")
			say line
			say "*" copies("-",70)
		end
	end
return

/* --- UpdatePlot --- */
UpdatePlot:
	call ps "showpage"
	/* Print Label */
	call psrgb 0.9,0.1,0.1
	call psfont "/Times-Italic-Bold",24
	call psmove 10,580
	call psshow "FlukaPlot"

	call psgray 0
	call psfont "/Helvetica",12
	call psmove 10,560
	call psshow date() time()
	call psrgb 0,0,0.5
	call psmove 10,540
	call psshow flukatitle

	y = 510; ys=13
	call psrgb 0,0,0.3
	call psmove 10,y; y=y-ys
	call psshow "Plot type:" type
	if type="DG" | type="PG" | type<>"D" then do
		call psmove 10,y; y=y-ys
		call psshow "Origin:" xx yy zz
		call psmove 10,y; y=y-ys
		call psshow "Extents:" extX extY
		call psmove 10,y; y=y-ys
		call psshow "Axis-X:" axisX
		call psmove 10,y; y=y-ys
		call psshow "Axis-Y:" axisY
	end
	if left(type,1)="D" | left(type,1)="P" then do
		call psmove 10,y; y=y-ys
		call psshow "Data file:" datafile
		call psmove 10,y; y=y-ys
		call psshow "Binning no:" binno
		if datatype(norm,"N") then
			invnorm="("1.0/norm")"
		else
			invnorm = ""
		call psmove 10,y; y=y-ys
		call psshow "Normalization:" norm invnorm
		call psmove 10,y; y=y-ys
		call psshow "Projection Axis:" projaxis
		call psmove 10,y; y=y-ys
		call psshow "Data limits:"
		call psmove 20,y; y=y-ys
		call psshow "X:" xmin xmax
		call psmove 20,y; y=y-ys
		call psshow "Y:" ymin ymax
		call psmove 20,y; y=y-ys
		call psshow "Z:" zmin zmax
		call psmove 20,y; y=y-ys
		call psshow "Plot limits:" plotmin plotmax
		call psmove 10,y; y=y-ys
		call psshow "Div. per decade:" divpd
	end

	call psmove 10,y; y=y-ys
	call psshow "Output file:" deffile

	call ps "save"
	call ps "-90 rotate -612 0 translate"
	f = open(deffile,"r")
	if f>0 then do
		do forever
			line = read(f)
			if eof(f) then leave
			/*if left(line,1)="%" then iterate*/
			if line="showpage" then iterate
			call ps line
		end
		call close f
	end; else
		say "+++ Cannot open file" deffile

	call ps "restore"
	call ps "setgray",1
	do i=100 to 200
		call psmove 10,i
		call psshow "Testing"
	end

	/* Flush and show */
	call psflush
return
