/* --- MCNP Library --- */
MCNPRectangular: procedure expose prim. surface
	parse arg n
	call MCNPAddSurf n,'+',"PX" prim.n.@xmin, "RPP X-Min"
	call MCNPAddSurf n,'-',"PX" prim.n.@xmax, "RPP X-Max"
	call MCNPAddSurf n,'+',"PY" prim.n.@ymin, "RPP Y-Min"
	call MCNPAddSurf n,'-',"PY" prim.n.@ymax, "RPP Y-Max"
	call MCNPAddSurf n,'+',"PZ" prim.n.@zmin, "RPP Z-Min"
	call MCNPAddSurf n,'-',"PZ" prim.n.@zmax, "RPP Z-Max"
return

MCNPBox: procedure expose prim. surface
	parse arg n

	P = prim.n.@P
	X = prim.n.@X
	Y = prim.n.@Y
	Z = prim.n.@Z

	/* outer corner */
	PO = VecAdd(P,VecAdd(X,VecAdd(Y,Z)))
	X = VecNorm(X)
	Y = VecNorm(Y)
	Z = VecNorm(Z)

	/* first 3 planes will be */
	call MCNPAddSurf n,'+',"P" X VecDot(P, X),"BOX X-Min"
	call MCNPAddSurf n,'+',"P" Y VecDot(P, Y),"BOX Y-Min"
	call MCNPAddSurf n,'+',"P" Z VecDot(P, Z),"BOX Z-Min"

	/* last 3 planes will be */
	call MCNPAddSurf n,'-',"P" X VecDot(PO, X),"BOX X-Max"
	call MCNPAddSurf n,'-',"P" Y VecDot(PO, Y),"BOX Y-Max"
	call MCNPAddSurf n,'-',"P" Z VecDot(PO, Z),"BOX Z-Max"
return

MCNPSphere: procedure expose prim. surface
	parse arg n

	parse var prim.n.@P x y z
	select
		when x=0 & y=0 & z=0 then
			call MCNPAddSurf n,'-',"SO" prim.n.@R,"SPH"
		when y=0 & z=0 then
			call MCNPAddSurf n,'-',"SX" X prim.n.@R,"SPH"
		when x=0 & z=0 then
			call MCNPAddSurf n,'-',"SY" Y prim.n.@R,"SPH"
		when x=0 & y=0 then
			call MCNPAddSurf n,'-',"SZ" Z prim.n.@R,"SPH"
		otherwise
			call MCNPAddSurf n,'-',"S" X Y Z prim.n.@R,"SPH"
	end
return

MCNPCylinder: procedure expose prim. surface transform
	parse arg n

	P = prim.n.@P
	Z = prim.n.@Z
	Radius = prim.n.@R
	Ptop = VecAdd(P,Z)
	Z = VecNorm(Z)

	call MCNPAddSurf n,'-',"CZ" Radius,"RCC cylinder",P "6j" Z
	call MCNPAddSurf n,'+',"P" Z VecDot(Z,P),"RCC base"
	call MCNPAddSurf n,'-',"P" Z VecDot(Z,Ptop),"RCC top"
return

MCNPInfCylinder: procedure expose prim. surface
	parse arg n,axis

	parse var prim.n.@P x y z
	R = prim.n.@R

	select
		when axis=='X' then do
			if x=0 & y=0 then
				call MCNPAddSurf n,'-',"CX" R,"CXX"
			else
				call MCNPAddSurf n,'-',"C/X" y z R,"CXX"
		end
		when axis=='Y' then do
			if z=0 & x=0 then
				call MCNPAddSurf n,'-',"CY" R,"CYY"
			else
				call MCNPAddSurf n,'-',"C/Y" x z R,"CYY"
		end
		when axis=='Z' then do
			if x=0 & y=0 then
				call MCNPAddSurf n,'-',"CZ" R,"CZZ"
			else
				call MCNPAddSurf n,'-',"C/Z" x y R,"CZZ"
		end
	end
return

MCNPInfEllipticalCylinder: procedure expose prim. surface
	parse arg n,axis

	parse var prim.n.@R a b c
	if a<>0 then a = 1/a**2
	if b<>0 then b = 1/b**2
	if c<>0 then c = 1/c**2

	call MCNPAddSurf n,'-',"SQ" a b c 0 0 0 (-1) prim.n.@P,"RE"axis" cylinder"
return

MCNPEllipticalCylinder: procedure expose prim. surface transform
	parse arg n

	P = prim.n.@P
	X = prim.n.@X
	Y = prim.n.@Y
	Z = prim.n.@Z

	Ptop = VecAdd(P,Z)

	RadiusX = VecLen(X)
	RadiusY = VecLen(Y)
	H  = VecLen(Z)

	/* normalize the coordinates */
	Zn = VecCross(X,Y)
	if VecDot(Z,Zn)<0 then do
		Z = VecCross(Y,X)
		Y = VecCross(X,Z)
	end; else do
		Z = Zn
		Y = VecCross(Z,X)
	end

	X = VecNorm(X)
	Y = VecNorm(Y)
	Z = VecNorm(Z)

	a = 1/(RadiusX**2)
	b = 1/(RadiusY**2)

	call MCNPAddSurf n,'-',"SQ" a b 0 0 0 0 (-1) 0 0 0,"REC cylinder",P X Y "3j"
	call MCNPAddSurf n,'+',"P" Z VecDot(Z,P),"REC base"
	call MCNPAddSurf n,'-',"P" Z VecDot(Z,Ptop),"REC top"
return

MCNPTruncatedCone: procedure expose prim. surface transform
	parse arg n

	P = prim.n.@P
	Z = prim.n.@Z
	apexRadius = prim.n.@aR
	baseRadius = prim.n.@bR

	Len  = VecLen(Z)
	Ptop = VecAdd(P,Z)
	Z = VecNorm(Z)
	S = (apexRadius-baseRadius)/Len

	call MCNPAddSurf n,'-',"KZ" (-baseRadius/S) S**2,"TRC",P "6j" Z
	call MCNPAddSurf n,'+',"P" Z VecDot(Z,P),"TRC base"
	call MCNPAddSurf n,'-',"P" Z VecDot(Z,Ptop),"TRC top"
return

MCNPPlane: procedure expose prim. surface
	parse arg n,axis

	select
		when axis=='X' then
			call MCNPAddSurf n,'-',"PX" VecX(prim.n.@P)

		when axis=='Y' then
			call MCNPAddSurf n,'-',"PY" VecY(prim.n.@P)

		when axis=='Z' then
			call MCNPAddSurf n,'-',"PZ" VecZ(prim.n.@P)

		otherwise
			P = prim.n.@P
			Normal = VecNorm(prim.n.@N)
			call MCNPAddSurf n,'-',"P" Normal VecDot(Normal, P)
	end
return

/* --- MCNPAddSurf --- */
MCNPAddSurf: procedure expose prim. surface transform
	parse arg n,op,surf,comment,tr
	s = prim.n.@surf.0

	if op = "+" then negop = "-"; else negop = "+"

	if s = "?" then do
		s = 1
		prim.n.@in = ""
		prim.n.@out = ""
	end; else
		s = s + 1

	prim.n.@surf.0 = s
	surface = surface + 1

	prim.n.@in = prim.n.@in op||surface
	if s=1 then
		prim.n.@out = negop||surface
	else
		prim.n.@out = prim.n.@out":"negop||surface

	if tr <> "" then do
		transform = transform + 1
		prim.n.@tr = tr
		trstr = transform
	end; else
		trstr = ""

	prim.n.@surf.s = trstr surf
return

/* --- MCNPCreateSurfaces --- */
MCNPCreateSurfaces: procedure expose prim. verbose
	surface = 0
	transform = 0
	if verbose then say "Creating the MCNP surfaces"

	do i=1 to prim.0
		primitive = prim.i.@prim
		if verbose then say "Creating primitive" i prim.i primitive
		select
			when primitive=="RPP" then call MCNPRectangular i
			when primitive=="BOX" then call MCNPBox i
			when primitive=="SPH" then call MCNPSphere i
			when primitive=="RCC" then call MCNPCylinder i
			when primitive=="REC" then call MCNPEllipticalCylinder i
			when primitive=="TRC" then call MCNPTruncatedCone i
			/**
			when primitive=="ELL" then call MCNPEllipsoid i
			when primitive=="WED" then call MCNPAngleWedge
			when primitive=="ARB" then call MCNPPolyhedron
			**/
			when primitive=="XYP" then call MCNPPlane i,'Z'
			when primitive=="XZP" then call MCNPPlane i,'Y'
			when primitive=="YZP" then call MCNPPlane i,'X'
			when primitive=="PLA" then call MCNPPlane i,'A'
			when primitive=="XCC" then call MCNPInfCylinder i,'X'
			when primitive=="YCC" then call MCNPInfCylinder i,'Y'
			when primitive=="ZCC" then call MCNPInfCylinder i,'Z'
			when primitive=="XEC" then call MCNPInfEllipticalCylinder i,'X'
			when primitive=="YEC" then call MCNPInfEllipticalCylinder i,'Y'
			when primitive=="ZEC" then call MCNPInfEllipticalCylinder i,'Z'
			otherwise
				call FLError "MCNP: unknown body type" primitive
		end
	end
return

/* --- MCNPWriteCells --- */
MCNPWriteCells: procedure expose verbose prim. region. material. what. card blackbody
	parse arg fout

	call MCNPWriteCommentLine "Cells"
        errors = 0

	do r=1 to region.0
		if verbose then
			say "Writing cell:" region.r

		/* construct the MCNP region line */
		if ^datatype(region.r.@mat,"NUM") then do
			Call FlError "No material was assigned for region" r region.r
			exit
		end
		mat = trunc(region.r.@mat)
		if material.mat.@density < 0 then
			line = "0"	/* create a void volume */
		else do
			density = -material.mat.@density
			line = region.r.@mat density
		end

		do or=1 to region.r.0
			if or>1 then line = line":"
			flukaline = region.r.or
			do while flukaline <> ""
				parse var flukaline op +1 id flukaline

				/* find number of surfaces composing id */
				oid = id
				if ^datatype(id,"N") then id = prim.id
				nsurf = prim.id.@surf.0
				if nsurf == "?" then do
					call lineout "<STDERR>","ERROR Primitive '"oid"' do not exist in region" r region.r
					call lineout "<STDERR>","Flukaline:" region.r.or
                                        errors = 1
					iterate
				end
				if nsurf>1 then do
					before="("
					after=")"
				end; else do
					before = ""
					after = ""
				end
				if op="+" then
					line = line,
					before||strip(prim.id.@in)||after
				else
					line = line,
					before||strip(prim.id.@out)||after
			end
		end
		/*line = changestr(": ",line,":")*/
		call MCNPWriteLine r,line,region.r.@comment,region.r.@line
	end
	call lineout fout
        if errors then call FlError "Abort due to errors.."
return

/* --- MCNPWriteSurfaces --- */
MCNPWriteSurfaces: procedure expose verbose prim.
	parse arg fout

	call MCNPWriteCommentLine "Surfaces"

	surface = 0
	do n=1 to prim.0
		nsurf = prim.n.@surf.0
		if nsurf == "?" then iterate
		do s=1 to nsurf
			surface = surface + 1
			if s=1 then
				call MCNPWriteLine surface,prim.n.@surf.s,,
					prim.n.@comment,prim.n.@line
			else
				call MCNPWriteLine surface,prim.n.@surf.s
		end
	end
	call lineout fout
return

/* --- MCNPWriteTransformations --- */
MCNPWriteTransformations: procedure expose verbose prim.
	parse arg fout

	call MCNPWriteCommentLine "Transformations"

	tr = 0
	do n=1 to prim.0
		if prim.n.@tr ^== "?" then do
			tr = tr + 1
			call MCNPWriteLine "TR"tr,prim.n.@tr,"Surface" n
		end
	end
return

/* --- MCNPWriteMaterials --- */
MCNPWriteMaterials: procedure expose verbose material.
	parse arg fout

	call MCNPWriteCommentLine "Materials"

	/* skip the black hole */
	do m=3 to material.0
		if material.m.@compound = "?" then iterate
		if ^material.m.@used then iterate
		mat = "M"right(m,3,0)
		call MCNPWriteLine mat,material.m.@compound,material.m
	end
return

/* --- MCNPWriteBiasing --- */
MCNPWriteBiasing: procedure expose verbose region.
	parse arg fout

	call MCNPWriteCommentLine "Biasing"

	line = ""
	repeat = 0
	prev = Bias(region.r.@bias)
	do r=2 to region.0-1
		if prev = Bias(region.r.@bias) then
			repeat = repeat + 1
		else do
			line = line prev
			if repeat=1 then line = line prev
			else
			if repeat>1 then line = line repeat"r"
			repeat = 0
			prev = Bias(region.r.@bias)
		end
	end
	line = line prev
	if repeat=1 then line = line prev
	else
	if repeat>1 then line = line repeat"r"

	line = line 0
	call MCNPWriteLine "IMP:n",line
return

Bias: if arg(1)=="?" then return 1; else return arg(1)

/* --- Write a MCNP line with comments and the original fluka line --- */
MCNPWriteLine: procedure expose fout
	parse arg cmd,args,comment,flukaline

	/* write first the comments */
	if comment <> "?" then
		call WriteMultiLine fout,changestr("09"x,comment," "),"C --- "

	/* then the original fluka line */
	call WriteMultiLine fout,flukaline,"C --- "

	/* parse the args not to exceed 60 chars */
	line = left(cmd,5)
	do while args <> ""
		do while args <> ""
			parse var args a args
			if length(line a)>72 then do
				args = a args
				leave
			end
			line = line a
		end
		call lineout fout,line
		line = "     "
	end
return

/* ---- MCNPWriteCommentLine ---- */
MCNPWriteCommentLine:
	call lineout fout,"C" center(arg(1),50,"-")
return
