/*
 * Library: povlib.r
 * Author: Vasilis.Vlachoudis@cern.ch 2004
 */

/* --- PovHeader --- */
PovHeader: procedure expose material.
	parse arg fout,fn "." .,overwritePovInc
	fninc = fn".inc"
	call lineout fout,'#include "colors.inc"'
	call lineout fout,'#include "textures.inc"'
	call lineout fout,'//#include "fluka.inc"'
	call lineout fout
	call lineout fout,'#include "'fninc'"'
	call PovInc fninc,overwritePovInc
return

/* --- PovName --- */
PovName: procedure
	name = translate(arg(1),,"-+*/:;<>,.&()[]{}!@#$%^&*=|","_")
	do until name=oldname
		oldname = name
		name = changestr("__",name,"_")
	end
return oldname

/* --- PovVec --- */
PovVec: procedure
	parse arg x y z,yn,zn
	if y="" then y=yn
	if z="" then z=zn
return "<"||(-x)","y","z">"

/* --- PovInc --- *
 * Generate the default include file with materials
 */
PovInc: procedure expose material. region.
	parse arg fn,overwrite
	f = open(fn,"r")	/* check if it exists */
	if f>0  then do
		call close f
		if overwrite=="" | overwrite=0 then return
	end
	f = open(fn,"w")

	/* Create a default & lights camera */
	call lineout f
	call lineout f,"// Ambient light"
	call lineout f,"global_settings{ambient_light Gray30}"
	call lineout f
	call lineout f,"sky_sphere {"
	call lineout f,"	pigment {"
	call lineout f,"		gradient y"
	call lineout f,"		color_map {"
	call lineout f,"			[0.000 0.002	color rgb <0.2, 0.2, 1.0>"
	call lineout f,"					color rgb <0.2, 0.2, 1.0>]"
	call lineout f,"			[0.002 0.200	color rgb <0.2, 0.1, 0.8>"
	call lineout f,"					color rgb <0.2, 0.2, 0.3>]"
	call lineout f,"		}"
	call lineout f,"		scale 2"
	call lineout f,"		translate -1"
	call lineout f,"	}"
	call lineout f,"	pigment {"
	call lineout f,"		bozo"
	call lineout f,"		turbulence 0.65"
	call lineout f,"		octaves 6"
	call lineout f,"		omega 0.7"
	call lineout f,"		lambda 2"
	call lineout f,"		color_map {"
	call lineout f,"			[ 0.0 0.1	color rgb <0.85, 0.85, 0.85>"
	call lineout f,"					color rgb <0.75, 0.75, 0.75>]"
	call lineout f,"			[ 0.1 0.5	color rgb <0.75, 0.75, 0.75>"
	call lineout f,"					color rgbt <1, 1, 1, 1>]"
	call lineout f,"			[ 0.5 1.0	color rgbt <1, 1, 1, 1>"
	call lineout f,"					color rgbt <1, 1, 1, 1>]"
	call lineout f,"		}"
	call lineout f,"		scale <0.2, 0.5, 0.2>"
	call lineout f,"	}"
	call lineout f,"	rotate -135*x"
	call lineout f,"}"

	call lineout f,"// Light sources"
	call lineout f,"light_source {"
	call lineout f,"	<120,240,100>"
	call lineout f,"	color White"
	call lineout f,"}"
	call lineout f
	call lineout f,"light_source {"
	call lineout f,"	<-100,240,50>"
	call lineout f,"	color Gray80"
	call lineout f,"	spotlight"
	call lineout f,"	radius 50"
	call lineout f,"	falloff 60"
	call lineout f,"	point_at <0,0,0>"
	call lineout f,"}"
	call lineout f
	call lineout f,"// Cameras"
	call lineout f,"camera {"
	call lineout f,"	location <100, 180, 250>"
	call lineout f,"	direction <0, 0, 1>"
	call lineout f,"	angle 40"
	call lineout f,"	up <0, 1, 0>"
	call lineout f,"	right <4/3, 0, 0>"
	call lineout f,"	look_at  <0, 0, 0>"
	call lineout f,"}"
	call lineout f

	call lineout f,"//---------------------- FLOOR --------------------------------"
	call lineout f,"plane { y, 0"
	call lineout f,"	texture {"
	call lineout f,"		pigment {"
	call lineout f,"			checker"
	call lineout f,"			color White"
	call lineout f,"			color Gray70"
	call lineout f,"			scale 20"
	call lineout f,"		}"
	call lineout f,"	}"
	call lineout f,"}"
	call lineout f,"//---------------------- Axis ---------------------------------"
	call lineout f,"#declare _axis=union {"
	call lineout f,"        cylinder { <0,0,0>,<0.9,0,0>,0.015 open }"
	call lineout f,"        cone { <0.9,0,0>,0.03,<1,0,0>,0 open }"
	call lineout f,"	texture {pigment {Candy_Cane scale 0.2} finish {ambient 0.15 diffuse 0.8}}"
	call lineout f,"	no_shadow"
	call lineout f,"}"
	call lineout f,"#local _axisScale=100;"
	call lineout f,"object {_axis rotate z*180 scale _axisScale} // X-Axis"
	call lineout f,"object {_axis rotate z*90  scale _axisScale} // Y-axis"
	call lineout f,"object {_axis rotate y*-90 scale _axisScale} // Z-axis"

	/* Colors definitions */
	c="DimGray DimGrey Gray Grey LightGray LightGrey VLightGray VLightGrey Aquamarine"
	c = c "BlueViolet Brown CadetBlue Coral CornflowerBlue DarkGreen DarkOliveGreen DarkOrchid"
	c = c "DarkSlateBlue DarkSlateGray DarkSlateGrey DarkTurquoise Firebrick ForestGreen"
	c = c "Gold Goldenrod GreenYellow IndianRed Khaki LightBlue LightSteelBlue LimeGreen"
	c = c "Maroon MediumAquamarine MediumBlue MediumForestGreen MediumGoldenrod"
	c = c "MediumOrchid MediumSeaGreen MediumSlateBlue MediumSpringGreen MediumTurquoise"
	c = c "MediumVioletRed MidnightBlue Navy NavyBlue Orange"
	c = c "OrangeRed Orchid PaleGreen Pink Plum"
	c = c "Salmon SeaGreen Sienna SkyBlue SlateBlue"
	c = c "SpringGreen SteelBlue Tan Thistle Turquoise"
	c = c "Violet VioletRed Wheat YellowGreen SummerSky"
	c = c "RichBlue Brass Copper Bronze Bronze2"
	c = c "Silver BrightGold OldGold Feldspar Quartz"
	c = c "Mica NeonPink DarkPurple NeonBlue CoolCopper"
	c = c "MandarinOrange LightWood MediumWood DarkWood SpicyPink"
	c = c "SemiSweetChoc BakersChoc Flesh NewTan NewMidnightBlue"
	c = c "VeryDarkBrown DarkBrown DarkTan GreenCopper DkGreenCopper"
	c = c "DustyRose HuntersGreen Scarlet Med_Purple Light_Purple Very_Light_Purple"
	Ncolors = words(c)

	CR  = "0A"x

	do i=1 to material.0
		if ^material.i.@used then iterate
		if material.i.@density<material.@povDensity then iterate

		name = PovName(material.i)
		upper name
		select
			when abbrev(name,"WATER") | name="H2O" then color="Water"
			when abbrev(name,"ALUMINUM")	then color="Aluminum"
			when abbrev(name,"GOLD")	then color="Gold_Metal"
			when abbrev(name,"SILVER")	then color="Silver_Metal"
			when abbrev(name,"IRON")	then color="Aluminum"
			when abbrev(name,"STEEL")	then color="Chrome_Metal"
			when abbrev(name,"COPPER")	then color="Copper_Metal"
			when abbrev(name,"CARBON")	then color="Copper_Metal"
			when abbrev(name,"GLASS")	then color="Glass"
			when abbrev(name,"CONCRETE")	then color="texture {"CR ||,
					"	pigment {color <0.8,0.8,0.7>}"CR ||,
					"	normal { bumps 0.2 scale 0.2 }"CR ||,
					"	finish { ambient 0.2 diffuse 0.6 phong .75 }}"
			when abbrev(name,"MARBLE")	then color="White_Marble"
			otherwise
				/* Choose a random color */
				color = "texture {pigment {color "word(c,random(1,Ncolors))"}}"
		end
		call lineout f,"#declare mat_"PovName(material.i)"="color

		/* try to guess materials by name */
	end
	call close f
return

/* --- PovPrimitives --- */
PovPrimitives: procedure expose verbose prim. FarAway
	parse arg fout

	call lineout fout
	call lineout fout,"//" center(" Primitive objects ",70,"-")
	do n=1 to prim.0
		name = prim.n
		if datatype(name,"N") then name="p"name

		primitive = prim.n.@prim
		sdum = ""
		what. = ""

		call lineout fout,"//" n"." prim.n primitive
		call WriteMultiLine fout,prim.n.@comment,"//"
		call write fout,"#declare" name "= "
		T  = "09"x
		T2 = T||T
		T3 = T2||T
		T4 = T3||T
		select
			when primitive=="RPP" then
				call lineout fout,"box {"||,
					PovVec(prim.n.@xmin,prim.n.@ymin,prim.n.@zmin)",",
					PovVec(prim.n.@xmax,prim.n.@ymax,prim.n.@zmax)"}"

			when primitive=="BOX" then do
				lx = VecLen(prim.n.@X)
				ly = VecLen(prim.n.@Y)
				lz = VecLen(prim.n.@Z)
				nx = VecNorm(prim.n.@X)
				ny = VecNorm(prim.n.@Y)
				nz = VecNorm(prim.n.@Z)
				call lineout fout,"box {<0,0,0>,"PovVec(lx,ly,lz)
				call lineout fout,T"matrix <"VecX(nx)","VecX(ny)","VecX(nz)","
				call lineout fout,T2||VecY(nx)","VecY(ny)","VecY(nz)","
				call lineout fout,T2||VecZ(nx)","VecZ(ny)","VecZ(nz)","
				call lineout fout,T2||VecX(prim.n.@P)","VecY(prim.n.@P)","VecZ(prim.n.@P)"> }"
			end

			when primitive=="SPH" then
				call lineout fout,"sphere {"PovVec(prim.n.@P)"," prim.n.@R"}"

			when primitive=="RCC" then
				call lineout fout,"cylinder {"PovVec(prim.n.@P)",",
					PovVec(VecAdd(prim.n.@P,prim.n.@Z))","prim.n.@R"}"

			when primitive=="REC" then do
				a = 1/VecLen(prim.n.@X)**2
				b = 1/VecLen(prim.n.@Y)**2
				nx = VecNorm(prim.n.@X)
				ny = VecNorm(prim.n.@Y)
				nz = VecNorm(prim.n.@Z)
				call lineout fout,"intersection {"
				call lineout fout,T2"quadric {<"a","||b",0>,<0,0,0>,<0,0,0>,-1"
				call lineout fout,T3"matrix <"VecX(nx)","VecX(ny)","VecX(nz)","
				call lineout fout,T4||VecY(nx)","VecY(ny)","VecY(nz)","
				call lineout fout,T4||VecZ(nx)","VecZ(ny)","VecZ(nz)","
				call lineout fout,T4||VecX(prim.n.@P)","VecY(prim.n.@P)","VecZ(prim.n.@P)"> }"
				call lineout fout,T2"plane {"PovVec(nz)","VecDot(nz,prim.n.@P)" inverse}"
				Ptop = VecAdd(prim.n.@P,prim.n.@Z)
				call lineout fout,T2"plane {"PovVec(nz)","VecDot(nz,Ptop)"}"
				call lineout fout,T"}"
			end

			when primitive=="TRC" then do
				call lineout fout,"cone {"PovVec(prim.n.@P)","prim.n.@bR",",
					PovVec(VecAdd(prim.n.@P,prim.n.@Z))","prim.n.@aR"}"
			end

			/*
			when primitive=="ELL" then call WriteEllipsoid
			when primitive=="WED" then call WriteAngleWedge
			when primitive=="ARB" then call WritePolyhedron
			*/
			when primitive=="XYP" then
				call lineout fout,"plane {z," VecZ(prim.n.@P)"}"

			when primitive=="XZP" then
				call lineout fout,"plane {y," VecY(prim.n.@P)"}"

			when primitive=="YZP" then
				call lineout fout,"plane {x," (-VecX(prim.n.@P))" inverse}"

			when primitive=="PLA" then do
				Normal = VecNorm(prim.n.@N)
				call lineout fout,"plane {"PovVec(Normal)",",
					VecDot(Normal,prim.n.@P)"}"
			end

			when primitive=="XCC" then do
				base = PovVec(-FarAway,VecY(prim.n.@P),VecZ(prim.n.@P))
				cap  = PovVec( FarAway,VecY(prim.n.@P),VecZ(prim.n.@P))
				call lineout fout,"cylinder {"base","cap","prim.n.@R" open}"
			end

			when primitive=="YCC" then do
				base = PovVec(VecX(prim.n.@P),-FarAway,VecZ(prim.n.@P))
				cap  = PovVec(VecX(prim.n.@P), FarAway,VecZ(prim.n.@P))
				call lineout fout,"cylinder {"base","cap","prim.n.@R" open}"
			end

			when primitive=="ZCC" then do
				base = PovVec(VecX(prim.n.@P),VecY(prim.n.@P),-FarAway)
				cap  = PovVec(VecX(prim.n.@P),VecY(prim.n.@P), FarAway)
				call lineout fout,"cylinder {"base","cap","prim.n.@R" open}"
			end
			when primitive=="XEC" | primitive=="YEC" | primitive=="ZEC" then do
				parse var prim.n.@R a b c
				if a<>0 then a = 1/a**2
				if b<>0 then b = 1/b**2
				if c<>0 then c = 1/c**2
				parse var prim.n.@P xo yo zo
				xo = -xo
				g = -2*a*xo
				h = -2*b*yo
				i = -2*c*zo
				j = a*xo**2 + b*yo**2 + c*zo**2 - 1
				call lineout fout,"quadric {<"a","||b","c">,<0,0,0>,<"g","||h","i">,"j"}"
			end
			otherwise
				call lineout fout,"Unknown object" primitive
		end
	end
return

/* --- PovRegions --- */
PovRegions: procedure expose verbose region. prim. material.
	parse arg fout

	call lineout fout
	call lineout fout
	call lineout fout,"//" center(" Regions ",70,"-")
	do r=1 to region.0
		call lineout fout
		m = region.r.@mat
		call lineout fout,"//" r"." "Rgn:"region.r "Mat:"material.m
		call WriteMultiLine fout,region.r.@comment,"//"
		if material.m.@density<material.@povDensity then iterate
		if region.r.0>1 then call lineout fout,"union {"

		finish = "texture {mat_"PovName(material.m)"}"

		do i=1 to region.r.0
			product = region.r.i
			Nproduct = words(product)
			if Nproduct>1 then call lineout fout,"	intersection {"
			do j=1 to Nproduct
				parse value word(product,j) with sign +1 prim
				if sign="-" then
					inverse=" inverse"
				else
					inverse = ""

				if datatype(prim,"N") then do
					prim = prim.prim
					if datatype(prim,"N") then prim="p"prim
				end
				if Nproduct>1 | region.r.0>1 then
					call lineout fout,"		object {"prim||inverse"}"
				else do
					call lineout fout,"		object {"prim||inverse
					call lineout fout,"			"finish
					call lineout fout,"		}"
				end
			end
			if Nproduct>1 then do
				if region.r.0=1 then
					call lineout fout,"		"finish
				call lineout fout,"	}"
			end
		end
		if region.r.0>1 then do
			call lineout fout,"	"finish
			call lineout fout,"} // union"
		end
	end
return
